@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -104,6 +104,27 @@ static void timers_update(struct ofono_stk *stk);
 		result.additional_len = sizeof(addn_info);	\
 		result.additional = addn_info;			\
 
+static gboolean convert_to_phone_number_format(const char *input_str,
+							char *output_str)
+{
+	char *digit;
+	char *digit_from = "01234567890abcABC";
+	char *digit_to = "01234567890*#p*#p";
+	int pos;
+
+	for (pos = 0; input_str[pos] != '\0'; pos++) {
+		digit = strchr(digit_from, input_str[pos]);
+		if (digit == NULL)
+			return FALSE;
+
+		output_str[pos] = digit_to[digit - digit_from];
+	}
+
+	output_str[pos] = '\0';
+
+	return TRUE;
+}
+
 static int stk_respond(struct ofono_stk *stk, struct stk_response *rsp,
 			ofono_stk_generic_cb_t cb)
 {
@@ -128,6 +149,7 @@ static int stk_respond(struct ofono_stk *stk, struct stk_response *rsp,
 	stk_command_free(stk->pending_cmd);
 	stk->pending_cmd = NULL;
 	stk->cancel_cmd = NULL;
+	stk->respond_on_exit = FALSE;
 
 	stk->driver->terminal_response(stk, tlv_len, tlv, cb, stk);
 
@@ -380,6 +402,11 @@ static struct stk_menu *stk_menu_create(const char *title,
 		if (item_icon_ids && item_icon_ids->len)
 			ret->items[i].icon_id = item_icon_ids->list[i];
 
+		if (ret->items[i].icon_id != 0 &&
+				item_icon_ids->qualifier ==
+				STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY)
+			ret->items[i].text[0]='\0';
+
 		if (item->id == default_id)
 			ret->default_item = i;
 	}
@@ -473,19 +500,36 @@ static void emit_menu_changed(struct ofono_stk *stk)
 	g_dbus_send_message(conn, signal);
 }
 
+static void cancel_pending_dtmf(struct ofono_stk *stk)
+{
+	struct ofono_voicecall *vc;
+
+	vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL,
+				__ofono_atom_get_modem(stk->atom));
+
+	if (vc) /* Should be always true here */
+		__ofono_voicecall_tone_cancel(vc, stk->dtmf_id);
+}
+
 static void user_termination_cb(enum stk_agent_result result, void *user_data)
 {
 	struct ofono_stk *stk = user_data;
 
-	if (result == STK_AGENT_RESULT_TERMINATE) {
-		stk->respond_on_exit = FALSE;
-		send_simple_response(stk, STK_RESULT_TYPE_USER_TERMINATED);
+	if (result != STK_AGENT_RESULT_TERMINATE)
+		return;
+
+	switch (stk->pending_cmd->type) {
+	case STK_COMMAND_TYPE_SEND_DTMF:
+		cancel_pending_dtmf(stk);
+		break;
 	}
+
+	send_simple_response(stk, STK_RESULT_TYPE_USER_TERMINATED);
 }
 
-static void stk_alpha_id_set(struct ofono_stk *stk,
-		const char *text, const struct stk_text_attribute *attr,
-		const struct stk_icon_id *icon)
+static gboolean stk_alpha_id_set(struct ofono_stk *stk,
+			const char *text, const struct stk_text_attribute *attr,
+			const struct stk_icon_id *icon)
 {
 	char *alpha = dbus_apply_text_attributes(text, attr);
 
@@ -495,7 +539,22 @@ static void stk_alpha_id_set(struct ofono_stk *stk,
 	 * better idea is found out.
 	 */
 	if (alpha == NULL)
-		return;
+		return FALSE;
+
+	if (stk->current_agent == NULL)
+		return FALSE;
+
+	/*
+	 * According to 3GPP TS 102.223 section 8.31:
+	 * If icon is self-explanatory, it replaces the alpha identifier or
+	 * text string.
+	 * If icon is not self-explanatory, it shall be displayed together
+	 * with the alpha identifier or text string.
+	 */
+
+	if (icon->id != 0 && icon->qualifier ==
+			STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY)
+		alpha[0]='\0';
 
 	if (stk->respond_on_exit)
 		stk_agent_display_action(stk->current_agent, alpha, icon,
@@ -504,11 +563,18 @@ static void stk_alpha_id_set(struct ofono_stk *stk,
 		stk_agent_display_action_info(stk->current_agent, alpha, icon);
 
 	g_free(alpha);
+
+	return TRUE;
 }
 
 static void stk_alpha_id_unset(struct ofono_stk *stk)
 {
-	stk_agent_request_cancel(stk->current_agent);
+	/*
+	 * If there is no default agent, then current agent also will be NULL.
+	 * So, call request cancel only when there is a valid current agent.
+	 */
+	if (stk->current_agent)
+		stk_agent_request_cancel(stk->current_agent);
 }
 
 static int duration_to_msecs(const struct stk_duration *duration)
@@ -551,7 +617,7 @@ static DBusMessage *stk_get_properties(DBusConnection *conn,
 					OFONO_PROPERTIES_ARRAY_SIGNATURE,
 					&dict);
 
-	str = stk->idle_mode_text;
+	str = stk->idle_mode_text ? stk->idle_mode_text : "";
 	ofono_dbus_dict_append(&dict, "IdleModeText", DBUS_TYPE_STRING, &str);
 
 	icon = stk->idle_mode_icon.id;
@@ -598,7 +664,6 @@ static void default_agent_notify(gpointer user_data)
 
 	stk->default_agent = NULL;
 	stk->current_agent = stk->session_agent;
-	stk->respond_on_exit = FALSE;
 }
 
 static void session_agent_notify(gpointer user_data)
@@ -617,7 +682,6 @@ static void session_agent_notify(gpointer user_data)
 
 	stk->session_agent = NULL;
 	stk->current_agent = stk->default_agent;
-	stk->respond_on_exit = FALSE;
 
 	if (stk->remove_agent_source) {
 		g_source_remove(stk->remove_agent_source);
@@ -786,27 +850,34 @@ static DBusMessage *stk_select_item(DBusConnection *conn,
 
 	DBG("");
 
-	if (stk_send_envelope(stk, &e, menu_selection_envelope_cb, 0))
-		return __ofono_error_failed(msg);
-
 	stk->pending = dbus_message_ref(msg);
 
+	if (stk_send_envelope(stk, &e, menu_selection_envelope_cb, 0))
+		__ofono_dbus_pending_reply(&stk->pending,
+					__ofono_error_failed(stk->pending));
+
 	return NULL;
 }
 
-static GDBusMethodTable stk_methods[] = {
-	{ "GetProperties",		"",	"a{sv}",stk_get_properties },
-	{ "SelectItem",			"yo",	"",	stk_select_item,
-					G_DBUS_METHOD_FLAG_ASYNC },
-	{ "RegisterAgent",		"o",	"",	stk_register_agent },
-	{ "UnregisterAgent",		"o",	"",	stk_unregister_agent },
-
+static const GDBusMethodTable stk_methods[] = {
+	{ GDBUS_METHOD("GetProperties",
+			NULL, GDBUS_ARGS({ "properties", "a{sv}" }),
+			stk_get_properties) },
+	{ GDBUS_ASYNC_METHOD("SelectItem",
+			GDBUS_ARGS({ "item", "y" }, { "agent", "o" }), NULL,
+			stk_select_item) },
+	{ GDBUS_METHOD("RegisterAgent",
+			GDBUS_ARGS({ "path", "o" }), NULL,
+			stk_register_agent) },
+	{ GDBUS_METHOD("UnregisterAgent",
+			GDBUS_ARGS({ "path", "o" }), NULL,
+			stk_unregister_agent) },
 	{ }
 };
 
-static GDBusSignalTable stk_signals[] = {
-	{ "PropertyChanged",	"sv" },
-
+static const GDBusSignalTable stk_signals[] = {
+	{ GDBUS_SIGNAL("PropertyChanged",
+			GDBUS_ARGS({ "name", "s" }, { "value", "v" })) },
 	{ }
 };
 
@@ -863,20 +934,17 @@ static gboolean handle_command_send_sms(const struct stk_command *cmd,
 					struct ofono_stk *stk)
 {
 	struct ofono_modem *modem = __ofono_atom_get_modem(stk->atom);
-	struct ofono_atom *sms_atom;
 	struct ofono_sms *sms;
 	GSList msg_list;
 	struct ofono_uuid uuid;
 
-	sms_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_SMS);
+	sms = __ofono_atom_find(OFONO_ATOM_TYPE_SMS, modem);
 
-	if (sms_atom == NULL || !__ofono_atom_get_registered(sms_atom)) {
+	if (sms == NULL) {
 		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
 		return TRUE;
 	}
 
-	sms = __ofono_atom_get_data(sms_atom);
-
 	extern_req_start(stk);
 
 	msg_list.data = (void *) &cmd->send_sms.gsm_sms;
@@ -907,11 +975,11 @@ static gboolean handle_command_set_idle_text(const struct stk_command *cmd,
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *path = __ofono_atom_get_path(stk->atom);
+	const struct stk_command_setup_idle_mode_text *sim =
+						&cmd->setup_idle_mode_text;
 	char *idle_mode_text;
 
-	idle_mode_text = dbus_apply_text_attributes(
-					cmd->setup_idle_mode_text.text,
-					&cmd->setup_idle_mode_text.text_attr);
+	idle_mode_text = dbus_apply_text_attributes(sim->text, &sim->text_attr);
 
 	if (idle_mode_text == NULL) {
 		rsp->result.type = STK_RESULT_TYPE_DATA_NOT_UNDERSTOOD;
@@ -921,6 +989,10 @@ static gboolean handle_command_set_idle_text(const struct stk_command *cmd,
 	if (stk->idle_mode_text)
 		g_free(stk->idle_mode_text);
 
+	if (sim->icon_id.id != 0 && sim->icon_id.qualifier ==
+			STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY)
+		idle_mode_text[0]='\0';
+
 	stk->idle_mode_text = idle_mode_text;
 
 	ofono_dbus_signal_property_changed(conn, path, OFONO_STK_INTERFACE,
@@ -928,8 +1000,8 @@ static gboolean handle_command_set_idle_text(const struct stk_command *cmd,
 						DBUS_TYPE_STRING,
 						&idle_mode_text);
 
-	if (stk->idle_mode_icon.id != cmd->setup_idle_mode_text.icon_id.id) {
-		memcpy(&stk->idle_mode_icon, &cmd->setup_idle_mode_text.icon_id,
+	if (stk->idle_mode_icon.id != sim->icon_id.id) {
+		memcpy(&stk->idle_mode_icon, &sim->icon_id,
 				sizeof(stk->idle_mode_icon));
 
 		ofono_dbus_signal_property_changed(conn, path,
@@ -1159,8 +1231,6 @@ static void request_selection_cb(enum stk_agent_result result, uint8_t id,
 {
 	struct ofono_stk *stk = user_data;
 
-	stk->respond_on_exit = FALSE;
-
 	switch (result) {
 	case STK_AGENT_RESULT_OK:
 	{
@@ -1212,7 +1282,7 @@ static gboolean handle_command_select_item(const struct stk_command *cmd,
 					stk->select_item_menu,
 					request_selection_cb, stk,
 					request_selection_destroy,
-					stk->timeout * 1000) < 0) {
+					stk->short_timeout * 1000) < 0) {
 		unsigned char no_cause_result[] = { 0x00 };
 
 		request_selection_destroy(stk);
@@ -1243,8 +1313,6 @@ static void display_text_cb(enum stk_agent_result result, void *user_data)
 	static unsigned char screen_busy_result[] = { 0x01 };
 	static struct ofono_error error = { .type = OFONO_ERROR_TYPE_FAILURE };
 
-	stk->respond_on_exit = FALSE;
-
 	/*
 	 * There are four possible paths for DisplayText with immediate
 	 * response flag set:
@@ -1321,11 +1389,18 @@ static gboolean handle_command_display_text(const struct stk_command *cmd,
 	}
 
 	if (qualifier & (1 << 7))
-		timeout = stk->timeout * 1000;
+		timeout = stk->short_timeout * 1000;
 
 	if (dt->duration.interval)
 		timeout = duration_to_msecs(&dt->duration);
 
+	if (cmd->display_text.immediate_response)
+		timeout = stk->timeout * 1000;
+
+	if (dt->icon_id.id != 0 && dt->icon_id.qualifier ==
+			STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY)
+		text[0]='\0';
+
 	err = stk_agent_display_text(stk->current_agent, text, &dt->icon_id,
 					priority, display_text_cb, stk,
 					display_text_destroy, timeout);
@@ -1386,8 +1461,6 @@ static void request_confirmation_cb(enum stk_agent_result result,
 	struct stk_command_get_inkey *cmd = &stk->pending_cmd->get_inkey;
 	struct stk_response rsp;
 
-	stk->respond_on_exit = FALSE;
-
 	switch (result) {
 	case STK_AGENT_RESULT_OK:
 		memset(&rsp, 0, sizeof(rsp));
@@ -1442,8 +1515,6 @@ static void request_key_cb(enum stk_agent_result result, char *string,
 	struct stk_command_get_inkey *cmd = &stk->pending_cmd->get_inkey;
 	struct stk_response rsp;
 
-	stk->respond_on_exit = FALSE;
-
 	switch (result) {
 	case STK_AGENT_RESULT_OK:
 		memset(&rsp, 0, sizeof(rsp));
@@ -1493,17 +1564,15 @@ static gboolean handle_command_get_inkey(const struct stk_command *cmd,
 						struct stk_response *rsp,
 						struct ofono_stk *stk)
 {
-	int timeout = stk->timeout * 1000;
+	int timeout = stk->short_timeout * 1000;
 	const struct stk_command_get_inkey *gi = &cmd->get_inkey;
 	char *text = dbus_apply_text_attributes(gi->text, &gi->text_attr);
 	uint8_t qualifier = stk->pending_cmd->qualifier;
 	gboolean alphabet = (qualifier & (1 << 0)) != 0;
 	gboolean ucs2 = (qualifier & (1 << 1)) != 0;
 	gboolean yesno = (qualifier & (1 << 2)) != 0;
-	/*
-	 * Note: immediate response and help parameter values are not
-	 * provided by current api.
-	 */
+	gboolean immediate = (qualifier & (1 << 3)) != 0;
+	/* Note: help parameter value is not provided by current api. */
 	int err;
 
 	if (text == NULL) {
@@ -1516,6 +1585,10 @@ static gboolean handle_command_get_inkey(const struct stk_command *cmd,
 
 	gettimeofday(&stk->get_inkey_start_ts, NULL);
 
+	if (gi->icon_id.id != 0 && gi->icon_id.qualifier ==
+			STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY)
+		text[0]='\0';
+
 	if (yesno)
 		err = stk_agent_request_confirmation(stk->current_agent,
 							text, &gi->icon_id,
@@ -1526,6 +1599,11 @@ static gboolean handle_command_get_inkey(const struct stk_command *cmd,
 						&gi->icon_id, ucs2,
 						request_key_cb, stk, NULL,
 						timeout);
+	else if (immediate)
+		err = stk_agent_request_quick_digit(stk->current_agent,
+							text, &gi->icon_id,
+							request_key_cb, stk,
+							NULL, timeout);
 	else
 		err = stk_agent_request_digit(stk->current_agent, text,
 						&gi->icon_id, request_key_cb,
@@ -1560,8 +1638,6 @@ static void request_string_cb(enum stk_agent_result result, char *string,
 	gboolean packed = (qualifier & (1 << 3)) != 0;
 	struct stk_response rsp;
 
-	stk->respond_on_exit = FALSE;
-
 	switch (result) {
 	case STK_AGENT_RESULT_OK:
 		memset(&rsp, 0, sizeof(rsp));
@@ -1608,6 +1684,10 @@ static gboolean handle_command_get_input(const struct stk_command *cmd,
 		return TRUE;
 	}
 
+	if (gi->icon_id.id != 0 && gi->icon_id.qualifier ==
+			STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY)
+		text[0]='\0';
+
 	if (alphabet)
 		err = stk_agent_request_input(stk->current_agent, text,
 						&gi->icon_id, gi->default_text,
@@ -1672,14 +1752,10 @@ static void call_setup_connected(struct ofono_call *call, void *data)
 static void call_setup_cancel(struct ofono_stk *stk)
 {
 	struct ofono_voicecall *vc;
-	struct ofono_atom *vc_atom;
 
-	vc_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
-						OFONO_ATOM_TYPE_VOICECALL);
-	if (vc_atom == NULL)
-		return;
+	vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL,
+				__ofono_atom_get_modem(stk->atom));
 
-	vc = __ofono_atom_get_data(vc_atom);
 	if (vc)
 		__ofono_voicecall_dial_cancel(vc);
 }
@@ -1694,13 +1770,12 @@ static void confirm_call_cb(enum stk_agent_result result, gboolean confirm,
 	static unsigned char busy_on_call_result[] = { 0x02 };
 	static unsigned char no_cause_result[] = { 0x00 };
 	char *alpha_id = NULL;
-	struct ofono_voicecall *vc = NULL;
-	struct ofono_atom *vc_atom;
+	struct ofono_voicecall *vc;
 	struct stk_response rsp;
+	char number[256];
+	char *pause_chr;
 	int err;
 
-	stk->respond_on_exit = FALSE;
-
 	switch (result) {
 	case STK_AGENT_RESULT_TIMEOUT:
 		confirm = FALSE;
@@ -1719,11 +1794,8 @@ static void confirm_call_cb(enum stk_agent_result result, gboolean confirm,
 		return;
 	}
 
-	vc_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
-						OFONO_ATOM_TYPE_VOICECALL);
-	if (vc_atom)
-		vc = __ofono_atom_get_data(vc_atom);
-
+	vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL,
+				__ofono_atom_get_modem(stk->atom));
 	if (vc == NULL) {
 		send_simple_response(stk, STK_RESULT_TYPE_NOT_CAPABLE);
 		return;
@@ -1739,7 +1811,20 @@ static void confirm_call_cb(enum stk_agent_result result, gboolean confirm,
 		}
 	}
 
-	err = __ofono_voicecall_dial(vc, sc->addr.number, sc->addr.ton_npi,
+	/* Convert the setup call number to phone number format */
+	if (convert_to_phone_number_format(sc->addr.number, number) == FALSE) {
+		send_simple_response(stk, STK_RESULT_TYPE_DATA_NOT_UNDERSTOOD);
+		return;
+	}
+
+	/* Remove the DTMF string from the phone number */
+	pause_chr = strchr(number, 'p');
+
+	if (pause_chr)
+		number[pause_chr - number] = '\0';
+
+	/* TODO: send the DTMF after call is connected */
+	err = __ofono_voicecall_dial(vc, number, sc->addr.ton_npi,
 					alpha_id, sc->icon_id_call_setup.id,
 					qualifier >> 1, call_setup_connected,
 					stk);
@@ -1778,6 +1863,54 @@ static void confirm_call_cb(enum stk_agent_result result, gboolean confirm,
 		stk_command_cb(&error, stk);
 }
 
+static void confirm_handled_call_cb(enum stk_agent_result result,
+					gboolean confirm, void *user_data)
+{
+	struct ofono_stk *stk = user_data;
+	const struct stk_command_setup_call *sc =
+					&stk->pending_cmd->setup_call;
+	struct ofono_voicecall *vc;
+	char number[256];
+	char *pause_chr;
+
+	if (stk->driver->user_confirmation == NULL)
+		goto out;
+
+	if (result != STK_AGENT_RESULT_OK) {
+		stk->driver->user_confirmation(stk, FALSE);
+		goto out;
+	}
+
+	if (convert_to_phone_number_format(sc->addr.number, number) == FALSE) {
+		stk->driver->user_confirmation(stk, FALSE);
+		goto out;
+	}
+
+	/* Remove the DTMF string from the phone number */
+	pause_chr = strchr(number, 'p');
+
+	if (pause_chr)
+		number[pause_chr - number] = '\0';
+
+	stk->driver->user_confirmation(stk, confirm);
+
+	vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL,
+				__ofono_atom_get_modem(stk->atom));
+	if (vc == NULL)
+		goto out;
+
+	__ofono_voicecall_set_alpha_and_icon_id(vc, number,
+						sc->addr.ton_npi,
+						sc->alpha_id_call_setup,
+						sc->icon_id_call_setup.id);
+
+	return;
+
+out:
+	stk_command_free(stk->pending_cmd);
+	stk->pending_cmd = NULL;
+}
+
 static gboolean handle_command_set_up_call(const struct stk_command *cmd,
 						struct stk_response *rsp,
 						struct ofono_stk *stk)
@@ -1787,7 +1920,6 @@ static gboolean handle_command_set_up_call(const struct stk_command *cmd,
 	static unsigned char busy_on_call_result[] = { 0x02 };
 	char *alpha_id = NULL;
 	struct ofono_voicecall *vc = NULL;
-	struct ofono_atom *vc_atom;
 	int err;
 
 	if (qualifier > 5) {
@@ -1804,11 +1936,8 @@ static gboolean handle_command_set_up_call(const struct stk_command *cmd,
 		return TRUE;
 	}
 
-	vc_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
-						OFONO_ATOM_TYPE_VOICECALL);
-	if (vc_atom)
-		vc = __ofono_atom_get_data(vc_atom);
-
+	vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL,
+				__ofono_atom_get_modem(stk->atom));
 	if (vc == NULL) {
 		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
 		return TRUE;
@@ -1854,14 +1983,9 @@ static gboolean handle_command_set_up_call(const struct stk_command *cmd,
 static void send_ussd_cancel(struct ofono_stk *stk)
 {
 	struct ofono_ussd *ussd;
-	struct ofono_atom *atom;
 
-	atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
-						OFONO_ATOM_TYPE_USSD);
-	if (atom == NULL)
-		return;
-
-	ussd = __ofono_atom_get_data(atom);
+	ussd = __ofono_atom_find(OFONO_ATOM_TYPE_USSD,
+					__ofono_atom_get_modem(stk->atom));
 	if (ussd)
 		__ofono_ussd_initiate_cancel(ussd);
 
@@ -1962,17 +2086,15 @@ static gboolean handle_command_send_ussd(const struct stk_command *cmd,
 	struct ofono_modem *modem = __ofono_atom_get_modem(stk->atom);
 	static unsigned char busy_on_ss_result[] = { 0x03 };
 	static unsigned char busy_on_ussd_result[] = { 0x08 };
-	struct ofono_atom *atom;
 	struct ofono_ussd *ussd;
 	int err;
 
-	atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_USSD);
-	if (atom == NULL || !__ofono_atom_get_registered(atom)) {
+	ussd = __ofono_atom_find(OFONO_ATOM_TYPE_USSD, modem);
+	if (ussd == NULL) {
 		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
 		return TRUE;
 	}
 
-	ussd = __ofono_atom_get_data(atom);
 	if (__ofono_ussd_is_busy(ussd)) {
 		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
 					busy_on_ussd_result);
@@ -1990,11 +2112,6 @@ static gboolean handle_command_send_ussd(const struct stk_command *cmd,
 					cmd->send_ussd.ussd_string.len,
 					send_ussd_callback, stk);
 
-	if (err >= 0) {
-		stk->cancel_cmd = send_ussd_cancel;
-
-		return FALSE;
-	}
 
 	if (err == -ENOSYS) {
 		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
@@ -2007,6 +2124,20 @@ static gboolean handle_command_send_ussd(const struct stk_command *cmd,
 		return TRUE;
 	}
 
+	if (err < 0) {
+		unsigned char no_cause_result[] = { 0x00 };
+
+		/*
+		 * We most likely got an out of memory error, tell SIM
+		 * to retry
+		 */
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
+		return TRUE;
+	}
+
+	stk->cancel_cmd = send_ussd_cancel;
+
 	stk_alpha_id_set(stk, cmd->send_ussd.alpha_id,
 				&cmd->send_ussd.text_attr,
 				&cmd->send_ussd.icon_id);
@@ -2028,12 +2159,8 @@ static gboolean handle_command_refresh(const struct stk_command *cmd,
 					struct ofono_stk *stk)
 {
 	struct ofono_error failure = { .type = OFONO_ERROR_TYPE_FAILURE };
-	struct ofono_sim *sim = NULL;
-	struct ofono_atom *sim_atom;
-	struct ofono_ussd *ussd = NULL;
-	struct ofono_atom *ussd_atom;
-	struct ofono_voicecall *vc = NULL;
-	struct ofono_atom *vc_atom;
+	struct ofono_modem *modem = __ofono_atom_get_modem(stk->atom);
+	struct ofono_sim *sim;
 	uint8_t addnl_info[1];
 	int err;
 	GSList *l;
@@ -2092,11 +2219,7 @@ static gboolean handle_command_refresh(const struct stk_command *cmd,
 					cmd->refresh.icon_id.qualifier);
 	DBG("Alpha ID: %s", cmd->refresh.alpha_id);
 
-	sim_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
-						OFONO_ATOM_TYPE_SIM);
-	if (sim_atom)
-		sim = __ofono_atom_get_data(sim_atom);
-
+	sim = __ofono_atom_find(OFONO_ATOM_TYPE_SIM, modem);
 	if (sim == NULL) {
 		if (rsp != NULL)
 			rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
@@ -2105,11 +2228,10 @@ static gboolean handle_command_refresh(const struct stk_command *cmd,
 	}
 
 	if (rsp != NULL) {
-		ussd_atom = __ofono_modem_find_atom(
-					__ofono_atom_get_modem(stk->atom),
-					OFONO_ATOM_TYPE_USSD);
-		if (ussd_atom)
-			ussd = __ofono_atom_get_data(ussd_atom);
+		struct ofono_ussd *ussd;
+		struct ofono_voicecall *vc;
+
+		ussd = __ofono_atom_find(OFONO_ATOM_TYPE_USSD, modem);
 
 		if (ussd && __ofono_ussd_is_busy(ussd)) {
 			addnl_info[0] = STK_RESULT_ADDNL_ME_PB_USSD_BUSY;
@@ -2120,11 +2242,7 @@ static gboolean handle_command_refresh(const struct stk_command *cmd,
 			return TRUE;
 		}
 
-		vc_atom = __ofono_modem_find_atom(
-					__ofono_atom_get_modem(stk->atom),
-					OFONO_ATOM_TYPE_VOICECALL);
-		if (vc_atom)
-			vc = __ofono_atom_get_data(vc_atom);
+		vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL, modem);
 
 		if (vc && __ofono_voicecall_is_busy(vc,
 					OFONO_VOICECALL_INTERACTION_NONE)) {
@@ -2200,7 +2318,7 @@ static gboolean handle_command_refresh(const struct stk_command *cmd,
 		g_slist_foreach(file_list, (GFunc) g_free, NULL);
 		g_slist_free(file_list);
 
-		return FALSE;
+		return TRUE;
 	}
 
 	rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
@@ -2260,39 +2378,69 @@ static gboolean handle_command_provide_local_info(const struct stk_command *cmd,
 				struct stk_response *rsp, struct ofono_stk *stk)
 {
 	switch (cmd->qualifier) {
+	case 0:
+		DBG("Unhandled - Local information according to current NAA");
+		break;
+	case 1:
+		DBG("Unhandled - IMEI");
+		break;
+	case 2:
+		DBG("Unhandled - Network measurement results");
+		break;
 	case 3:
 		DBG("Date, time and time zone");
 		get_time(rsp);
 		return TRUE;
-
 	case 4:
 		DBG("Language setting");
 		get_lang(rsp, stk);
 		return FALSE;
+	case 6:
+		DBG("Unhandled - Access Technology");
+		break;
+	case 7:
+		DBG("Unhandled - ESN of the terminal");
+		break;
+	case 8:
+		DBG("Unhandled - IMEISV of the terminal");
+		break;
+	case 9:
+		DBG("Unhandled - Search mode");
+		break;
+	case 10:
+		DBG("Unhandled - Charge state of Battery");
+		break;
+	case 11:
+		DBG("Unhandled - MEID");
+		break;
+	case 13:
+		DBG("Unhandled - Broadcast Network information");
+		break;
+	case 14:
+		DBG("Unhandled - Multiple Access Technologies");
+		break;
+	case 15:
+		DBG("Unhandled - Location information for multiple access"
+			" technologies");
+		break;
+	case 16:
+		DBG("Unhandled - Measurement results for multiple access"
+			" technologies");
+		break;
 
 	default:
 		ofono_info("Unsupported Provide Local Info qualifier: %d",
 				cmd->qualifier);
-		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
-		return TRUE;
+		break;
 	}
+
+	rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
+	return TRUE;
 }
 
 static void send_dtmf_cancel(struct ofono_stk *stk)
 {
-	struct ofono_voicecall *vc = NULL;
-	struct ofono_atom *vc_atom;
-
-	stk->respond_on_exit = FALSE;
-
-	vc_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
-						OFONO_ATOM_TYPE_VOICECALL);
-	if (vc_atom)
-		vc = __ofono_atom_get_data(vc_atom);
-
-	if (vc) /* Should be always true here */
-		__ofono_voicecall_tone_cancel(vc, stk->dtmf_id);
-
+	cancel_pending_dtmf(stk);
 	stk_alpha_id_unset(stk);
 }
 
@@ -2300,8 +2448,6 @@ static void dtmf_sent_cb(int error, void *user_data)
 {
 	struct ofono_stk *stk = user_data;
 
-	stk->respond_on_exit = FALSE;
-
 	stk_alpha_id_unset(stk);
 
 	if (error == ENOENT) {
@@ -2334,35 +2480,23 @@ static gboolean handle_command_send_dtmf(const struct stk_command *cmd,
 {
 	static unsigned char not_in_speech_call_result[] = { 0x07 };
 	struct ofono_voicecall *vc = NULL;
-	struct ofono_atom *vc_atom;
-	char dtmf[256], *digit;
-	char *dtmf_from = "01234567890abcABC";
-	char *dtmf_to = "01234567890*#p*#p";
-	int err, pos;
-
-	vc_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
-						OFONO_ATOM_TYPE_VOICECALL);
-	if (vc_atom)
-		vc = __ofono_atom_get_data(vc_atom);
+	char dtmf[256];
+	int err;
 
+	vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL,
+				__ofono_atom_get_modem(stk->atom));
 	if (vc == NULL) {
 		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
 		return TRUE;
 	}
 
 	/* Convert the DTMF string to phone number format */
-	for (pos = 0; cmd->send_dtmf.dtmf[pos] != '\0'; pos++) {
-		digit = strchr(dtmf_from, cmd->send_dtmf.dtmf[pos]);
-		if (digit == NULL) {
-			rsp->result.type = STK_RESULT_TYPE_DATA_NOT_UNDERSTOOD;
-			return TRUE;
-		}
-
-		dtmf[pos] = dtmf_to[digit - dtmf_from];
+	if (convert_to_phone_number_format(cmd->send_dtmf.dtmf, dtmf) ==
+						FALSE) {
+		rsp->result.type = STK_RESULT_TYPE_DATA_NOT_UNDERSTOOD;
+		return TRUE;
 	}
 
-	dtmf[pos] = '\0';
-
 	err = __ofono_voicecall_tone_send(vc, dtmf, dtmf_sent_cb, stk);
 
 	if (err == -ENOSYS) {
@@ -2413,8 +2547,6 @@ static void play_tone_cb(enum stk_agent_result result, void *user_data)
 {
 	struct ofono_stk *stk = user_data;
 
-	stk->respond_on_exit = FALSE;
-
 	switch (result) {
 	case STK_AGENT_RESULT_OK:
 	case STK_AGENT_RESULT_TIMEOUT:
@@ -2435,7 +2567,7 @@ static gboolean handle_command_play_tone(const struct stk_command *cmd,
 	static const struct {
 		const char *name;
 		/* Continuous true/false according to 02.40 */
-		gboolean continuous;
+		gboolean repeatable;
 	} tone_infos[] = {
 		/* Default */
 		[0x00] = { "general-beep", FALSE },
@@ -2445,9 +2577,9 @@ static gboolean handle_command_play_tone(const struct stk_command *cmd,
 		[0x02] = { "busy", TRUE },
 		[0x03] = { "congestion", TRUE },
 		[0x04] = { "radio-path-acknowledge", FALSE },
-		[0x05] = { "radio-path-not-available", FALSE },
+		[0x05] = { "radio-path-not-available", TRUE },
 		[0x06] = { "error", TRUE },
-		[0x07] = { "call-waiting", FALSE },
+		[0x07] = { "call-waiting", TRUE },
 		[0x08] = { "ringing-tone", TRUE },
 
 		/* Proprietary */
@@ -2504,7 +2636,19 @@ static gboolean handle_command_play_tone(const struct stk_command *cmd,
 	else
 		timeout = manufacturer_timeout;
 
-	if (!tone_infos[pt->tone].continuous)
+	if (pt->icon_id.id != 0 && pt->icon_id.qualifier ==
+			STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY)
+		text[0]='\0';
+
+	/*
+	 * According to TS 102.223 section 6.6.6:
+	 * "the length of time for which the Terminal shall generate the tone,
+	 * if the tone is contunious or repeatable.  For single tones, the
+	 * value of this data object shall be ignored by the Terminal.  If no
+	 * duration is specified, the Terminal shall default to a duration
+	 * determined by the Terminal manufacturer
+	 */
+	if (!tone_infos[pt->tone].repeatable)
 		/* Duration ignored */
 		err = stk_agent_play_tone(stk->current_agent, text,
 						&pt->icon_id, vibrate,
@@ -2547,8 +2691,6 @@ static void confirm_launch_browser_cb(enum stk_agent_result result,
 	struct ofono_error failure = { .type = OFONO_ERROR_TYPE_FAILURE };
 	struct stk_response rsp;
 
-	stk->respond_on_exit = FALSE;
-
 	switch (result) {
 	case STK_AGENT_RESULT_TIMEOUT:
 		confirm = FALSE;
@@ -2612,6 +2754,53 @@ static gboolean handle_command_launch_browser(const struct stk_command *cmd,
 	return FALSE;
 }
 
+static void setup_call_handled_cancel(struct ofono_stk *stk)
+{
+	struct ofono_voicecall *vc;
+
+	vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL,
+				__ofono_atom_get_modem(stk->atom));
+
+	if (vc != NULL)
+		__ofono_voicecall_clear_alpha_and_icon_id(vc);
+}
+
+static gboolean handle_setup_call_confirmation_req(struct stk_command *cmd,
+						struct ofono_stk *stk)
+{
+	const struct stk_command_setup_call *sc = &cmd->setup_call;
+	int err;
+	char *alpha_id = dbus_apply_text_attributes(
+					sc->alpha_id_usr_cfm ?
+					sc->alpha_id_usr_cfm : "",
+					&sc->text_attr_usr_cfm);
+	if (alpha_id == NULL)
+		goto out;
+
+	if (stk->current_agent == FALSE)
+		goto out;
+
+	err = stk_agent_confirm_call(stk->current_agent, alpha_id,
+					&sc->icon_id_usr_cfm,
+					confirm_handled_call_cb,
+					stk, NULL,
+					stk->timeout * 1000);
+	g_free(alpha_id);
+
+	if (err < 0)
+		goto out;
+
+	stk->cancel_cmd = setup_call_handled_cancel;
+
+	return TRUE;
+
+out:
+	if (stk->driver->user_confirmation)
+		stk->driver->user_confirmation(stk, FALSE);
+
+	return FALSE;
+}
+
 static void stk_proactive_command_cancel(struct ofono_stk *stk)
 {
 	if (stk->immediate_response)
@@ -2698,6 +2887,9 @@ void ofono_stk_proactive_command_notify(struct ofono_stk *stk,
 		case STK_COMMAND_TYPE_GET_INPUT:
 		case STK_COMMAND_TYPE_PLAY_TONE:
 		case STK_COMMAND_TYPE_SETUP_CALL:
+		case STK_COMMAND_TYPE_SEND_SMS:
+		case STK_COMMAND_TYPE_SEND_USSD:
+		case STK_COMMAND_TYPE_SEND_DTMF:
 			send_simple_response(stk, STK_RESULT_TYPE_NOT_CAPABLE);
 			return;
 
@@ -2818,12 +3010,23 @@ void ofono_stk_proactive_command_notify(struct ofono_stk *stk,
 		stk_command_cb(&error, stk);
 }
 
+static gboolean handled_alpha_id_set(struct ofono_stk *stk,
+			const char *text, const struct stk_text_attribute *attr,
+			const struct stk_icon_id *icon)
+{
+	if (stk_alpha_id_set(stk, text, attr, icon) == FALSE)
+		return FALSE;
+
+	stk->cancel_cmd = stk_alpha_id_unset;
+	return TRUE;
+}
+
 void ofono_stk_proactive_command_handled_notify(struct ofono_stk *stk,
 						int length,
 						const unsigned char *pdu)
 {
-	struct stk_command *cmd;
 	struct stk_response dummyrsp;
+	gboolean ok = FALSE;
 
 	/*
 	 * Modems send us the proactive command details and terminal responses
@@ -2832,69 +3035,75 @@ void ofono_stk_proactive_command_handled_notify(struct ofono_stk *stk,
 	 * responses here
 	 */
 	if (length > 0 && pdu[0] == 0x81) {
-		stk_alpha_id_unset(stk);
+		stk_proactive_command_cancel(stk);
 		return;
 	}
 
 	stk_proactive_command_cancel(stk);
 
-	cmd = stk_command_new_from_pdu(pdu, length);
-
-	if (cmd == NULL || cmd->status != STK_PARSE_RESULT_OK) {
-		ofono_error("Can't parse proactive command");
-
-		if (cmd)
-			stk_command_free(cmd);
+	stk->pending_cmd = stk_command_new_from_pdu(pdu, length);
+	if (stk->pending_cmd == NULL)
 		return;
+
+	if (stk->pending_cmd->status != STK_PARSE_RESULT_OK) {
+		ofono_error("Can't parse modem-handled proactive command");
+		ok = FALSE;
+		goto out;
 	}
 
-	DBG("type: %d", cmd->type);
-
-	switch (cmd->type) {
-	case STK_COMMAND_TYPE_MORE_TIME:
-		break;
+	DBG("type: %d", stk->pending_cmd->type);
 
+	switch (stk->pending_cmd->type) {
 	case STK_COMMAND_TYPE_SEND_SMS:
-		stk_alpha_id_set(stk, cmd->send_sms.alpha_id,
-				&cmd->send_sms.text_attr,
-				&cmd->send_sms.icon_id);
+		ok = handled_alpha_id_set(stk,
+					stk->pending_cmd->send_sms.alpha_id,
+					&stk->pending_cmd->send_sms.text_attr,
+					&stk->pending_cmd->send_sms.icon_id);
 		break;
 
 	case STK_COMMAND_TYPE_SETUP_IDLE_MODE_TEXT:
-		handle_command_set_idle_text(cmd, &dummyrsp, stk);
+		handle_command_set_idle_text(stk->pending_cmd, &dummyrsp, stk);
 		break;
 
 	case STK_COMMAND_TYPE_SETUP_MENU:
-		handle_command_set_up_menu(cmd, &dummyrsp, stk);
+		handle_command_set_up_menu(stk->pending_cmd, &dummyrsp, stk);
 		break;
 
 	case STK_COMMAND_TYPE_SETUP_CALL:
-		/* TODO */
+		ok = handle_setup_call_confirmation_req(stk->pending_cmd, stk);
 		break;
 
 	case STK_COMMAND_TYPE_SEND_USSD:
-		stk_alpha_id_set(stk, cmd->send_ussd.alpha_id,
-				&cmd->send_ussd.text_attr,
-				&cmd->send_ussd.icon_id);
+		ok = handled_alpha_id_set(stk,
+					stk->pending_cmd->send_ussd.alpha_id,
+					&stk->pending_cmd->send_ussd.text_attr,
+					&stk->pending_cmd->send_ussd.icon_id);
 		break;
 
 	case STK_COMMAND_TYPE_SEND_SS:
-		stk_alpha_id_set(stk, cmd->send_ss.alpha_id,
-				&cmd->send_ss.text_attr,
-				&cmd->send_ss.icon_id);
+		ok = handled_alpha_id_set(stk,
+					stk->pending_cmd->send_ss.alpha_id,
+					&stk->pending_cmd->send_ss.text_attr,
+					&stk->pending_cmd->send_ss.icon_id);
+		break;
 
 	case STK_COMMAND_TYPE_SEND_DTMF:
-		stk_alpha_id_set(stk, cmd->send_dtmf.alpha_id,
-				&cmd->send_dtmf.text_attr,
-				&cmd->send_dtmf.icon_id);
+		ok = handled_alpha_id_set(stk,
+					stk->pending_cmd->send_dtmf.alpha_id,
+					&stk->pending_cmd->send_dtmf.text_attr,
+					&stk->pending_cmd->send_dtmf.icon_id);
 		break;
 
 	case STK_COMMAND_TYPE_REFRESH:
-		handle_command_refresh(stk->pending_cmd, NULL, stk);
+		ok = handle_command_refresh(stk->pending_cmd, NULL, stk);
 		break;
 	}
 
-	stk_command_free(cmd);
+out:
+	if (ok == FALSE) {
+		stk_command_free(stk->pending_cmd);
+		stk->pending_cmd = NULL;
+	}
 }
 
 int ofono_stk_driver_register(const struct ofono_stk_driver *d)
@@ -3025,7 +3234,7 @@ void ofono_stk_register(struct ofono_stk *stk)
 	__ofono_atom_register(stk->atom, stk_unregister);
 
 	stk->timeout = 180; /* 3 minutes */
-	stk->short_timeout = 20; /* 20 seconds */
+	stk->short_timeout = 25; /* 25 seconds */
 	stk->envelope_q = g_queue_new();
 }
 
