@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -38,6 +38,10 @@
 #include "stkutil.h"
 #include "stkagent.h"
 
+#ifndef DBUS_TIMEOUT_INFINITE
+#define DBUS_TIMEOUT_INFINITE ((int) 0x7fffffff)
+#endif
+
 enum allowed_error {
 	ALLOWED_ERROR_GO_BACK	= 0x1,
 	ALLOWED_ERROR_TERMINATE	= 0x2,
@@ -55,6 +59,9 @@ struct stk_agent {
 	DBusPendingCall *call;
 	void *user_cb;
 	void *user_data;
+	int min_length;
+	int max_length;
+	ofono_bool_t hidden_entry;
 	ofono_destroy_func user_destroy;
 
 	const struct stk_menu *request_selection_menu;
@@ -535,14 +542,24 @@ static void get_digit_cb(DBusPendingCall *call, void *data)
 
 	if (dbus_message_get_args(reply, NULL,
 					DBUS_TYPE_STRING, &digit,
-					DBUS_TYPE_INVALID) == FALSE ||
-			strlen(digit) != 1 ||
-			!valid_phone_number_format(digit)) {
+					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to GetDigit()");
 		remove_agent = TRUE;
 		goto error;
 	}
 
+	if (strlen(digit) != 1 || !strspn(digit, "0123456789*#+")) {
+		ofono_error("Invalid character");
+		remove_agent = TRUE;
+		goto error;
+	}
+
+	if (agent->hidden_entry && digit[0] == '+') {
+		ofono_error("The character + is not allowed in this mode");
+		remove_agent = TRUE;
+		goto error;
+	}
+
 	cb(result, digit, agent->user_data);
 
 	CALLBACK_END();
@@ -574,6 +591,40 @@ int stk_agent_request_digit(struct stk_agent *agent, const char *text,
 	agent->user_cb = cb;
 	agent->user_data = user_data;
 	agent->user_destroy = destroy;
+	agent->hidden_entry = FALSE;
+
+	dbus_pending_call_set_notify(agent->call, get_digit_cb, agent, NULL);
+
+	return 0;
+}
+
+int stk_agent_request_quick_digit(struct stk_agent *agent, const char *text,
+					const struct stk_icon_id *icon,
+					stk_agent_string_cb cb, void *user_data,
+					ofono_destroy_func destroy, int timeout)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+
+	agent->msg = dbus_message_new_method_call(agent->bus, agent->path,
+							OFONO_SIM_APP_INTERFACE,
+							"RequestQuickDigit");
+	if (agent->msg == NULL)
+		return -ENOMEM;
+
+	dbus_message_append_args(agent->msg,
+					DBUS_TYPE_STRING, &text,
+					DBUS_TYPE_BYTE, &icon->id,
+					DBUS_TYPE_INVALID);
+
+	if (dbus_connection_send_with_reply(conn, agent->msg, &agent->call,
+						timeout) == FALSE ||
+			agent->call == NULL)
+		return -EIO;
+
+	agent->user_cb = cb;
+	agent->user_data = user_data;
+	agent->user_destroy = destroy;
+	agent->hidden_entry = TRUE;
 
 	dbus_pending_call_set_notify(agent->call, get_digit_cb, agent, NULL);
 
@@ -656,6 +707,7 @@ static void get_digits_cb(DBusPendingCall *call, void *data)
 	enum stk_agent_result result;
 	gboolean remove_agent;
 	char *string;
+	int len, span;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE,
@@ -677,6 +729,25 @@ static void get_digits_cb(DBusPendingCall *call, void *data)
 		goto error;
 	}
 
+	len = strlen(string);
+
+	if (len < agent->min_length || len > agent->max_length) {
+		ofono_error("Length not acceptable");
+		remove_agent = TRUE;
+		goto error;
+	}
+
+	if (agent->hidden_entry)
+		span = strspn(string, "0123456789*#");
+	else
+		span = strspn(string, "0123456789*#+");
+
+	if (span != len) {
+		ofono_error("Invalid character found");
+		remove_agent = TRUE;
+		goto error;
+	}
+
 	cb(result, string, agent->user_data);
 
 	CALLBACK_END();
@@ -720,6 +791,9 @@ int stk_agent_request_digits(struct stk_agent *agent, const char *text,
 	agent->user_cb = cb;
 	agent->user_data = user_data;
 	agent->user_destroy = destroy;
+	agent->min_length = min_val;
+	agent->max_length = max_val;
+	agent->hidden_entry = hidden_val;
 
 	dbus_pending_call_set_notify(agent->call, get_digits_cb, agent, NULL);
 
@@ -734,6 +808,7 @@ static void get_input_cb(DBusPendingCall *call, void *data)
 	enum stk_agent_result result;
 	gboolean remove_agent;
 	char *string;
+	int len;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE,
@@ -755,6 +830,14 @@ static void get_input_cb(DBusPendingCall *call, void *data)
 		goto error;
 	}
 
+	len = g_utf8_strlen(string, -1);
+
+	if (len < agent->min_length || len > agent->max_length) {
+		ofono_error("Length not acceptable");
+		remove_agent = TRUE;
+		goto error;
+	}
+
 	cb(result, string, agent->user_data);
 
 	CALLBACK_END();
@@ -799,6 +882,9 @@ int stk_agent_request_input(struct stk_agent *agent, const char *text,
 	agent->user_cb = cb;
 	agent->user_data = user_data;
 	agent->user_destroy = destroy;
+	agent->min_length = min_val;
+	agent->max_length = max_val;
+	agent->hidden_entry = hidden_val;
 
 	dbus_pending_call_set_notify(agent->call, get_input_cb, agent, NULL);
 
@@ -1005,7 +1091,7 @@ int stk_agent_display_action_info(struct stk_agent *agent, const char *text,
 					DBUS_TYPE_INVALID);
 
 	if (dbus_connection_send_with_reply(conn, agent->msg, &agent->call,
-						0) == FALSE ||
+					DBUS_TIMEOUT_INFINITE) == FALSE ||
 			agent->call == NULL)
 		return -EIO;
 
@@ -1132,8 +1218,8 @@ int stk_agent_display_action(struct stk_agent *agent,
 					DBUS_TYPE_INVALID);
 
 	if (dbus_connection_send_with_reply(conn, agent->msg, &agent->call,
-						0) == FALSE ||
-						agent->call == NULL)
+					DBUS_TIMEOUT_INFINITE) == FALSE ||
+			agent->call == NULL)
 		return -EIO;
 
 	agent->user_cb = cb;
